#!/usr/bin/env python

# MIT License
#
# Copyright (c) 2018 Sam Kovaka <skovaka@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys                         
import os
import argparse
import numpy as np
import uncalled as unc

def get_parser(conf):
    parser = argparse.ArgumentParser(
            description="Rapidly maps raw nanopore signal to DNA references", 
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    sp = parser.add_subparsers(dest="subcmd")

    index_parser = sp.add_parser(
            "index", 
            help="Builds the UNCALLED index of a FASTA reference", 
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    add_index_opts(index_parser, conf)

    map_parser = sp.add_parser(
            "map", 
            help="Map fast5 files to a DNA reference", 
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    add_bwa_opt(map_parser, conf)
    add_fast5_opts(map_parser, conf)
    add_map_opts(map_parser, conf)

    rt_parser = sp.add_parser(
            "realtime", 
            help="Perform real-time targeted (ReadUntil) sequencing",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    add_bwa_opt(rt_parser, conf)
    add_map_opts(rt_parser, conf)
    add_ru_opts(rt_parser, conf)
    add_realtime_opts(rt_parser, conf)

    sim_parser = sp.add_parser(
            "sim", 
            help="Simulate real-time target sequencing.", 
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    add_bwa_opt(sim_parser, conf)
    add_sim_opts(sim_parser, conf)
    add_map_opts(sim_parser, conf)
    add_ru_opts(sim_parser, conf)

    ps_parser = sp.add_parser(
            "pafstats", 
            help="Computes speed and accuracy of UNCALLED mappings.", #Given an UNCALLED PAF file, will compute mean/median BP mapped per second, number of BP required to map each read, and total number of milliseconds to map each read. Can also optionally compute accuracy with respect to reference alignments, for example output by minimap2.",
            formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    unc.pafstats.add_opts(ps_parser)
    #TODO move here

    #lp_parser = sp.add_parser("list-ports", help="List the port of all MinION devices detected in the current MinKNOW session", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    #add_list_ports_opts(lp_parser, conf)

    return parser

def add_index_opts(p, conf):
    p.add_argument(
            "fasta_filename", 
            type=str, 
            help="FASTA file to index"
    )
    p.add_argument(
            "-o", "--bwa-prefix", 
            type=str, default=None, 
            help="Index output prefix. Will use input fasta filename by default"
    )
    p.add_argument(
            "-s", "--max-sample-dist", 
            type=int, default=100, 
            help="Maximum average sampling distance between reference alignments."
    )
    p.add_argument(
            "--min-samples", 
            type=int, default=50000, 
            help="Minimum number of alignments to produce (approximate, due to deterministically random start locations)"
    )
    p.add_argument(
            "--max-samples", 
            type=int, default=1000000, 
            help="Maximum number of alignments to produce (approximate, due to deterministically random start locations)"
    )
    p.add_argument(
            "-k", "--kmer-len", 
            type=int, default=5,
            help="Model k-mer length"
    )
    p.add_argument(
            "-1", "--matchpr1", 
            default=0.6334, type=float, 
            help="Minimum event match probability"
    )
    p.add_argument(
            "-2", "--matchpr2", 
            type=float, default=0.9838, 
            help="Maximum event match probability"
    )
    p.add_argument(
            "-f", "--pathlen-percentile", 
            type=float, default=0.05, 
            help=""
    )
    p.add_argument(
            "-m", "--max-replen", 
            type=int, default=100, 
            help=""
    )
    p.add_argument(
            "--probs", 
            type=str, default=None, 
            help="Find parameters with specified target probabilites (comma separated)"
    )
    p.add_argument(
            "--speeds", 
            type=str, default=None, 
            help="Find parameters with specified speed coefficents (comma separated)"
    )

def add_bwa_opt(p, conf):
    p.add_argument(
            "bwa_prefix", 
            type=str, 
            help="BWA prefix to mapping to. Must be processed by \"uncalled index\"."
    )
    p.add_argument(
            "-p", "--idx-preset", 
            type=str, default=conf.idx_preset, 
            help="Mapping mode"
    )

def add_ru_opts(p, conf):
    #TODO: selectively enrich or deplete refs in index
    p.add_argument(
            "-c", "--max-chunks", 
            type=int, default=conf.max_chunks, required=True, 
            help="Will give up on a read after this many chunks have been processed. Only has effect when --unblock is set"
    )
    p.add_argument(
            "--chunk-time", 
            type=float, default=1, required=False, 
            help="Length of chunks in seconds"
    )

    modes = p.add_mutually_exclusive_group(required=True)
    modes.add_argument(
            "-D", "--deplete", action='store_const', 
            const=unc.RealtimePool.DEPLETE, dest='realtime_mode', 
            help="Will eject reads that align to index"
    )
    modes.add_argument(
            "-E", "--enrich", action='store_const',
            const=unc.RealtimePool.ENRICH, dest='realtime_mode', 
            help="Will eject reads that don't align to index"
    )

    active = p.add_mutually_exclusive_group()
    active.add_argument(
            "--full", action='store_const', 
            const=unc.RealtimePool.FULL, dest='active_chs',
            help="Will monitor all pores if set (default)"
    )
    active.add_argument(
            "--even", action='store_const', 
            const=unc.RealtimePool.EVEN, dest='active_chs', 
            help="Will only monitor even pores if set"
    )
    active.add_argument(
            "--odd", action='store_const', 
            const=unc.RealtimePool.ODD, dest='active_chs', 
            help="Will only monitor odd pores if set")

def add_sim_opts(p, conf):
    p.add_argument(
            "fast5s", nargs='+', type=str, 
            help="Reads to unc. Can be a directory which will be recursively searched for all files with the \".fast5\" extension, a text file containing one fast5 filename per line, or a comma-separated list of fast5 file names."
    )
    p.add_argument(
            "-r", "--recursive", 
            action="store_true"
    )
    p.add_argument(
            "--ctl-seqsum", 
            type=str, required=True, 
            help=""
    )
    p.add_argument(
            "--unc-seqsum", 
            type=str, required=True, 
            help=""
    )
    p.add_argument(
            "--unc-paf", 
            type=str, required=True, 
            help=""
    )
    p.add_argument(
            "--sim-speed", 
            type=float, default=conf.sim_speed, 
            help="")

def add_realtime_opts(p, conf):
    p.add_argument(
            "--host", 
            type=str, default=conf.host, 
            help=unc.Conf.host.__doc__
    )
    p.add_argument(
            "--port", 
            type=int, default=conf.port, 
            help=unc.Conf.port.__doc__
    )
    p.add_argument(
            "--duration", 
            type=float, default=conf.duration, 
            help=unc.Conf.duration.__doc__
    )

#def add_list_ports_opts(p, conf):
#    p.add_argument(
#            "--log-dir", 
#            type=str, default="/var/log/MinKNOW", 
#            help="Directory to find MinKNOW log files"
#    )

def add_fast5_opts(p, conf):
    p.add_argument(
            "fast5s", nargs='+', type=str, 
            help="Reads to map. Can be a directory which will be recursively searched for all files with the \".fast5\" extension, a text file containing one fast5 filename per line, or a comma-separated list of fast5 file names."
    )
    p.add_argument(
            "-r", "--recursive", 
            action="store_true"
    )
    p.add_argument(
            "-l", "--read-list", 
            type=str, default=None, 
            help=unc.Conf.read_list.__doc__
    )
    p.add_argument(
            "-n", "--max-reads", 
            type=int, default=None, 
            help=unc.Conf.max_reads.__doc__
    )

#TODO get defautls from conf
def add_map_opts(p, conf):
    p.add_argument(
            "-t", "--threads", 
            type=int, default=conf.threads, 
            help="Number of threads to use for mapping"
    )
    p.add_argument(
            "--num-channels", 
            type=int, default=conf.num_channels, 
            help="Number of channels used in sequencing. If provided will use unique mapper for each channel. Useful for streaming normalization simulation."
    )
    p.add_argument(
            "-e", "--max-events", 
            type=int, default=conf.max_events, 
            help="Will give up on a read after this many events have been processed"
    )

def load_conf(argv):
    conf = unc.Conf()
    parser = get_parser(conf)
    args = parser.parse_args()

    for a, v in vars(args).items():
        if v == None:
            continue

        if (not a.startswith("_")):
            if hasattr(conf, a):
                setattr(conf, a, v)

    return conf, args



#!/usr/bin/env python

import sys
import argparse
import numpy as np
from uncalled.sim_utils import SeqsumProfile
from uncalled.pafstats import parse_paf

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Calculates enrichment from read until simulation")
    parser.add_argument("-u", "--uncalled-fname", required=True, type=str, default=None, help="Simulator output PAF file")
    parser.add_argument("-s", "--seq-sum", required=True, type=str, default=None, help="Control sequencing summary")
    parser.add_argument("-m", "--minimap-fname", required=True, type=str, default=None, help="Minimap2 PAF file of the control reads aligned to a reference containing the target (or off-target, in the case of depletion) sequences")
    parser.add_argument("-x", "--bwa-prefix", required=True, type=str, default=None, help="BWA reference used during the simulation")
    parser.add_argument("--deplete", action='store_true')
    parser.add_argument("--enrich", action='store_true')
    parser.add_argument("-t", "--sim-speed", required=False, type=float, default=1, help="Speed that the simulator was run at")
    args = parser.parse_args()



    if args.deplete == args.enrich:
        sys.stderr.write("Need to specify one\n")
        sys.exit(1)

    sys.stderr.write("Reading reference annotation\n")
    ann_in = open("%s.ann" % args.bwa_prefix)
    nrefs = int(ann_in.readline().split()[1])
    ref_seqs = set()
    for i in range(nrefs):
        ref_seqs.add(ann_in.readline().split()[1])
        ann_in.readline()

    sys.stderr.write("Reading uncalled\n")
    sim_duration = 0
    unc_reads = dict()
    for p in parse_paf(args.uncalled_fname):

        end = (p.tags['st'][0]/4000) + (p.qr_len/450)
        sim_duration = max(sim_duration, end)

        v = (p.qr_len, p.tags.get('ej',(None,0))[0], p.tags.get('dl', (0,0))[0])
        if p.qr_name in unc_reads:
            unc_reads[p.qr_name].append(v)
        else:
            unc_reads[p.qr_name] = [v]

    sys.stderr.write("Reading minimap2\n")

    mm2_maps = [(p.qr_name, p.rf_name)
                for p in parse_paf(args.minimap_fname) 
                if p.is_mapped and p.tags['tp'][0] == 'P']

    mapped_reads = {q for q,r in mm2_maps}

    tgt_reads = {q for q,r in mm2_maps
                 if (args.deplete and r not in ref_seqs) or
                    (not args.deplete and r in ref_seqs)}

    sys.stderr.write("Reading summary\n")

    read_info = dict()

    uc_counts = np.zeros((2,2))

    co = ct = 0
    uo = ut = 0

    ctl_reads = SeqsumProfile(args.seq_sum)
    ctl_reads.rm_scans()

    for i in range(len(ctl_reads)):

        read_id = ctl_reads.ids[i]
        start_time = ctl_reads.sts[i]
        duration = ctl_reads.lns[i]
        end_time = ctl_reads.ens[i]
        template_duration = ctl_reads.tds[i]
        seqlen = ctl_reads.bps[i]
        template_start = ctl_reads.tms[i]

        ontgt = read_id in tgt_reads
        
        if ontgt: ct += seqlen
        else: co += seqlen

        unc_alns = unc_reads.get(read_id, None)

        if unc_alns == None:
            continue

        bpps = seqlen/template_duration

        for unc_est,eject_time,delay_time in unc_alns:
            ejected = eject_time != None

            if ejected:
                unclen = bpps * ((unc_est/450.0) + (delay_time/4000.0) + eject_time - template_start)
                if ontgt: 
                    ut += min(seqlen, unclen)
                else:     
                    uo += min(seqlen, unclen)

            elif ontgt:
                ut += seqlen
            else:         
                uo += seqlen

    co /= 1e6 
    ct /= 1e6 
    uo /= 1e6 
    ut /= 1e6 

    sys.stderr.write("\n")
    print("unc_on_bp\t%.6f" % (ut/args.sim_speed))
    print("unc_total_bp\t%.6f" % ((ut+uo)/args.sim_speed))
    print("cnt_on_bp\t%.6f" % ct)
    print("cnt_total_bp\t%.6f" % (ct+co))




#!/usr/bin/env python

import sys, os
import numpy as np
import argparse
import bisect
from scipy.stats import linregress
from file_read_backwards import FileReadBackwards
from collections import defaultdict
import re

SEED_LEN = 22
MIN_CLUST = 25
SAMPLE_RATE = 4000
CHUNK_LEN = 4000

#Guppy basecalled event stride and kmer length
BCE_STRIDE = 5
BCE_K = 4

MAX_CHUNK_DEF = 3

#Cigar parsing
CIG_OPS_STR = "MIDNSHP=X"
CIG_RE = re.compile("(\d+)(["+CIG_OPS_STR+"])")
CIG_OPS = set(CIG_OPS_STR)
CIG_INCR_ALL = {'M','=', 'X'}
CIG_INCR_RD = CIG_INCR_ALL | {'I','S'}
CIG_INCR_RF = CIG_INCR_ALL | {'D','N'}

class DebugParser:
    CONF_PAD_COEF = 2

    def __init__(self, 
                 debug_prefix, 
                 unc_paf, 
                 mm2_paf=None,
                 bce_moves=None,
                 bwa_index=None,
                 min_chunk=None,
                 max_chunk=None,
                 min_samp=None,
                 max_samp=None,
                 tgt_cid=None,
                 max_path_fm=0,
                 load_seeds=True,
                 load_events=True,
                 load_paths=True,
                 load_conf=True):

        self.rid = unc_paf.qr_name

        #Cofident seed cluster ID, confident event, and reference length
        #(conf_evt is where mapping would normally end)
        self.conf_cid = unc_paf.tags.get('sc', (None,)*2)[0]
        self.conf_evt = unc_paf.tags.get('ce', (None,)*2)[0]
        self.conf_samp = None

        self.mm2_paf = mm2_paf

        self.idx = bwa_index

        #if self.tgt_cid is None:
        #    self.tgt_cid = self.conf_cid

        if min_samp is None and min_chunk is None:
            self.min_samp = self.min_chunk = 0
        elif min_chunk is None:
            self.min_samp = min_samp
            self.min_chunk = (min_samp-1) // CHUNK_LEN
        elif min_samp is None:
            self.min_samp = self.min_chunk * SAMPLE_RATE
            self.min_chunk = min_chunk
        else:
            self.min_samp = max(min_samp, min_chunk * SAMPLE_RATE)
            self.min_chunk = max(min_chunk, (min_samp-1) // CHUNK_LEN)


        if max_samp is None and max_chunk is None:
            if self.conf_evt is None:
                self.max_chunk = MAX_CHUNK_DEF
                self.max_samp = self.max_chunk*SAMPLE_RATE
            else:
                self.max_chunk = self.max_samp = None
        elif max_samp is None:
            self.max_chunk = max_chunk
            self.max_samp = max_chunk * SAMPLE_RATE
        elif max_chunk is None:
            self.max_chunk = max(0, (max_samp-1) // CHUNK_LEN)
            self.max_samp = max_samp
        else:
            self.max_chunk = max(max_chunk, (max_samp-1) // CHUNK_LEN)
            self.max_samp = max(max_samp, self.max_chunk * SAMPLE_RATE)

        self.min_evt = 0 if self.min_samp == 0 else None
        self.max_evt = None

        self.chunk_evt_bounds = [0]

        #TODO below should be w/r/t reference coordinates?
        #allow argument specified
        #or detect based on confident cluster (or just specify seed cluster)

        #TBD in parse_seeds
        self.conf_len = None 
        self.conf_clust = None
        self.max_clust = None

        self.max_clust_len = None

        if mm2_paf is not None:
            self.ref_name = mm2_paf.rf_name
            if mm2_paf.is_fwd:
                self.min_ref = mm2_paf.rf_st
            else:
                self.min_ref = -mm2_paf.rf_en
        else:
            self.min_ref = None
        self.max_ref = None

        #TODO clean these up
        self.conf_pbs = dict()
        self.conf_dots = set()
        self.path_lens = defaultdict(int)
        self.seed_kmers = dict()
        self.conf_ref_bounds = None

        self.evts_loaded = False
        if load_events:
            self.evts_in = open(debug_prefix + self.rid + "_events.tsv")
            self.parse_events()

        self.seeds_loaded = False
        if load_seeds:
            self.seeds_in  = open(debug_prefix + self.rid + "_seeds.bed")
            self.parse_seeds()

        self.bc_loaded = False
        if mm2_paf is not None and bce_moves is not None:
            self.parse_bc_aln(bce_moves)

        self.max_path_fm = max_path_fm
        if max_path_fm > 0:
            if self.idx is None:
                sys.stderr.write("Error: must include BWA index to include paths with FM overlaps\n")
                sys.exit(1)

            if self.conf_clust.fwd:
                st = self.min_ref
                en = self.max_ref
            else:
                st = -self.max_ref
                en = -self.min_ref

            fwd_fms, rev_fms = self.idx.range_to_fms(
                    self.ref_name, st, en
            )
            fms = fwd_fms if self.mm2_paf.is_fwd else rev_fms

            self.fm_to_ref = {fms[i] : i for i in range(len(fms))}
            self.range_fms = np.sort(fms)

        else:
            self.fwd_fms = self.rev_fms = None

        #TODO: paths
        self.paths_loaded = False
        if load_paths:
            self.paths_fname = debug_prefix + self.rid + "_paths.tsv"
            self.parse_paths()

        self.conf_loaded = False
        if load_conf:
            self.conf_fname = debug_prefix + self.rid + "_conf.tsv"
            self.parse_conf()

    def parse_events(self, incl_norm=True):
        if self.evts_loaded: return False
        self.evts_in.readline()

        self.events = list()
        self.norms = list()

        self.win_means = list()
        self.win_stdvs = list()
        self.evt_mask = list()

        self.evt_mask_map = list()

        unmask_evts = 0

        evt = 0
        chunk = 0
        next_chunk_samp = CHUNK_LEN

        for line in self.evts_in:
            tabs = line.split()
            st,ln,mask = map(int, tabs[:2] + tabs[-1:])
            mn,sd,norm_sc,norm_sh,win_mn,win_sd = map(float, tabs[2:-1])
            en = st+ln

            if st >= next_chunk_samp:
                chunk += 1
                next_chunk_samp += CHUNK_LEN
                self.chunk_evt_bounds.append(evt)

            if evt == self.conf_evt:
                self.conf_samp = st

                if self.max_samp is None:
                    self.max_samp = (chunk+1)*CHUNK_LEN-1
                    self.max_chunk = chunk

            if en < self.min_samp: 
                evt += 1
                continue

            if self.min_evt is None:
                print(en, self.min_samp)
                self.min_evt = evt
                print(self.min_evt, "MIN")

            if self.max_samp is not None and st > self.max_samp: 
                break
            
            self.events.append( (st,ln,mn,sd) )
            self.norms.append( (norm_sc,norm_sh) )

            self.win_means.append(win_mn)
            self.win_stdvs.append(win_sd)
            self.evt_mask.append(mask == 1)

            if self.evt_mask[-1]:
                self.evt_mask_map.append(evt)

            evt += 1

        if self.max_evt is None or self.max_evt >= evt: 
            self.max_evt = evt

        if self.max_samp is None:
            self.max_samp = st+ln
            self.max_chunk = (self.max_samp-1) // CHUNK_LEN

        self.evts_loaded = True

        self.win_means = np.array(self.win_means)
        self.win_stdvs = np.array(self.win_stdvs)
        self.evt_mask = np.array(self.evt_mask)
        self.evt_mask_map = np.array(self.evt_mask_map)

        return True
    
    def normed_event(self, e):
        scale,shift = self.norms[e]
        return scale*self.events[e][2]+shift

    def parse_seeds(self, expire_coef=None):
        if self.seeds_loaded: return False

        if expire_coef != None:
            SeedCluster.EXPIRE_COEF = expire_coef

        self.clusts = dict()
        self.clusts_exp = list()
    
        clust_ids = set()

        conf_clust = None
        for line in self.seeds_in:
            rf,st,en,name,strand = line.split()
            st,en,evt,pb,cid = map(int, [st,en]+name.split(":"))

            #handle expiration
            #should this (still) be here?
            clust = self.clusts.get(cid, None)
            replace = (
                clust is not None and 
                clust.expired(evt) and
                (cid != self.conf_cid or evt < self.conf_evt)
            )

            if clust == None or replace:
                if replace:
                    self.clusts_exp.append(clust)

                clust = SeedCluster(evt,rf,st,en,strand=="+",cid)
                self.clusts[cid] = clust

                clust_ids.add(cid)

            else:
                clust.add_seed(evt,rf,st,en)

            if clust.id == self.conf_cid:
                conf_clust = clust
                if clust.fwd:
                    #self.conf_pbs[(evt,pb)] = en
                    self.conf_pbs[(evt,pb)] = st
                else:
                    #self.conf_pbs[(evt,pb)] = -st
                    self.conf_pbs[(evt,pb)] = -en + 1

            if clust > self.max_clust:
                self.max_clust = clust

        self.mapped = conf_clust is not None

        #Handle unmapped reads
        #TODO: mark them more explicitly
        if not self.mapped:
            if self.max_clust is not None:
                conf_clust = self.max_clust
                self.conf_evt = conf_clust.evts[-1]
            else:
                self.conf_evt = self.max_evt

        self._set_conf_clust(conf_clust)
        self._parse_expired(clust_ids)

        self.seeds_loaded = True
        return True

    #sets confident cluster related vars
    def _set_conf_clust(self, cc):
        self.conf_clust = cc

        if cc is None: return

        self.conf_idx = np.searchsorted(cc.evts, self.conf_evt, side='right')-1
        self.conf_len = cc.lens[self.conf_idx]

        #TODO use cc.evrf_ratio(e,r)
        evt_span = self.conf_evt - cc.evt_st

        if self.max_evt is None:
            max_max_evt = self.conf_evt + (evt_span * self.CONF_PAD_COEF)
            self.max_evt = int(np.round(min(cc.evt_en+1, max_max_evt)))

        #TODO rename min/max_clust_ref
        self.min_idx = np.searchsorted(cc.evts, self.min_evt, side='left')
        self.max_idx = np.searchsorted(cc.evts, self.max_evt, side='right')-1

        if cc.fwd:
            rst = cc.rsts[self.min_idx] - SEED_LEN
            ren = cc.rens[self.max_idx]+1
        else:
            rst = -cc.rsts[self.min_idx] - SEED_LEN
            ren = -cc.rens[self.max_idx]-1

        ref_span = ren-rst
        if ref_span > 0:
            self.evrf_ratio = (self.max_evt) / ref_span
        else:
            self.evrf_ratio = 1

        self.max_clust_len = cc.lens[self.max_idx]

        if self.min_ref is None:
            self.min_ref = rst
        if self.max_ref is None:
            self.max_ref = ren
        if self.ref_name is None:
            self.ref_name = cc.ref


    #finds clusters that should expire
    #by the end of loaded events
    def _parse_expired(self, clust_ids):
        for cid in clust_ids:
            c = self.clusts[cid]
            if c.expired(self.max_evt):
                self.clusts_exp.append(c)
                del self.clusts[cid]

    def _parse_moves(self, moves):
        return np.flip()


    def parse_paths(self):
        if self.paths_loaded: return False

        path_counts = list()

        paths_fwd = open(self.paths_fname)
        head_tabs = paths_fwd.readline().split()
        C = {head_tabs[i] : i for i in range(len(head_tabs))}

        #paths_fwd.close()
        #paths_rev = FileReadBackwards(self.paths_fname)
        #for line in paths_rev:

        for line in paths_fwd:
            tabs = line.split()
            if tabs[0] == head_tabs[0]: continue
            path_id  = tuple(map(int, tabs[C['id']].split(':')))

            ref_st = self.conf_pbs.get(path_id, None)

            is_conf = ref_st is not None

            if not is_conf and self.max_path_fm == 0: continue

            evt, pb = path_id
            moves = [bool(c=='1') for c in tabs[C['moves']]]

            if evt < self.min_evt or evt >= self.max_evt: 
                continue

            e = evt-self.min_evt

            if not self.evt_mask[e-len(moves)]: 
                continue

            evts = np.arange(e-len(moves), e) + 1

            if not is_conf:
                fm_len    =   int(tabs[C['fm_len']])

                if fm_len > self.max_path_fm: continue

                fm_start  =   int(tabs[C['fm_start']])
                full_len  =   int(tabs[C['full_len']])

                i = np.searchsorted(self.range_fms, fm_start)

                while i < len(self.range_fms) and self.range_fms[i] < fm_start+fm_len: 
                    ref_en = self.fm_to_ref[self.range_fms[i]]
                    refs = ref_en - np.sum(moves) + np.cumsum(np.flip(moves)) - 3

                    print(ref_en, path_id)

                    for e,r in zip(evts,refs):
                        if e < 0: continue
                        self.path_lens[(e,r)] = max(self.path_lens[(e,r)], np.log2(fm_len))
                    i += 1

            else:
                refs = ref_st + (np.cumsum(np.flip(moves)) - 1)

                for e,r in zip(evts,refs):
                    if e < 0: continue
                    self.conf_dots.add( (e,r) )

            #kmer       =       tabs[C['kmer']]
            #match_prob = float(tabs[C['match_prob']])
            #self.seed_kmers[(e, refs[-1])] = (kmer, match_prob)


            #print(path_id,parent,fm_start,fm_len,kmer,full_len,seed_prob,moves)

        #    #Store number of paths at each event position
        #    while len(path_counts) - 1 < evt:
        #        path_counts.append(0)
        #    path_counts[-1] += 1

        #self.paths_loaded = True

        return True

    def parse_conf(self):
        if self.conf_loaded: return False

        self.conf_evts = list()
        self.conf_clusts = list()
        self.conf_tops = list()
        self.conf_means = list()

        conf_in = open(self.conf_fname)
        conf_in.readline()

        for line in conf_in:
            evt,clust,top,mean = line.split()
            self.conf_evts.append(int(evt))
            self.conf_clusts.append(int(clust))
            self.conf_tops.append(float(top))
            self.conf_means.append(float(mean))
            

    def parse_bc_aln(self, bce_moves):
        #List of ref coords for each query (read) cord
        qr_to_rfs = self._cig_query_to_refs(self.mm2_paf)

        #basecalled event start, end, and stride
        bce_samp_st, bce_moves_pac = bce_moves
            
        bce_samp_en = bce_samp_st + len(bce_moves) * BCE_STRIDE

        bce_st = 0
        bce_en = int((self.max_samp-bce_samp_st+1) // BCE_STRIDE)
        bce_moves = np.unpackbits(bce_moves_pac)[bce_st:bce_en]

        #Read coord of each basecalled event
        bce_qrs = np.cumsum(bce_moves)
        i = np.searchsorted(bce_qrs, self.mm2_paf.qr_st)
        
        #bce_evts = list()
        bce_samps = list()
        bce_refs = list()

        samp = bce_samp_st
        for qr in bce_qrs:
            if samp >= self.min_samp:
                for rf in qr_to_rfs[qr]:
                    bce_samps.append(samp)
                    bce_refs.append(rf)

            samp += BCE_STRIDE

        self.bce_samps = np.array(bce_samps)
        self.bce_refs = np.array(bce_refs) - BCE_K + 1
        self.max_ref = self.min_ref + max(bce_refs)


    def _cig_query_to_refs(self, paf):
        cig = paf.tags.get('cg', (None,)*2)[0]
        if cig is None: return None

        qr_rfs = defaultdict(list)

        qr_i = paf.qr_st
        rf_i = 0#paf.rf_st

        cig_ops = CIG_RE.findall(cig)

        if not paf.is_fwd:
            cig_ops = reversed(cig_ops)

        for l,c in cig_ops:
            l = int(l)
            incr_qr = c in CIG_INCR_RD
            incr_rf = c in CIG_INCR_RF
            qr_j = qr_i + (l if incr_qr else 1)
            rf_j = rf_i + (l if incr_rf else 1)
            for qr,rf in zip(range(qr_i, qr_j), range(rf_i, rf_j)):
                qr_rfs[qr].append(rf)

            if incr_qr:
                qr_i = qr_j 

            if incr_rf:
                rf_i = rf_j 

        return qr_rfs

class SeedCluster:
    EXPIRE_COEF = 5.5

    #TODO add start evt at st-SEED_LEN?
    #would make reference len easier to breakdown
    #and could be nice for plotting?
    def __init__(self, evt, rf, st, en, fwd, cid):
        self.id = cid
        self.rf = rf
        self.evts = [evt]
        self.blocks = [(st, en)] #TODO only use rsts?
        self.gains = [en-st]

        self.rsts = [st] #TODO consider strand
        self.rens = [en] #TODO consider strand

        self.lens = [self.gains[0]]
        self.fwd = fwd
        self.exp_evt = None

    def expired(self, evt=None):
        return False #

        #if evt == None:
        #    return self.exp_evt is not None

        #if evt - self.evt_en > self.ref_len * self.EXPIRE_COEF:
        #    if self.exp_evt is None:
        #        self.exp_evt = int(np.round(
        #            self.evt_en + (self.ref_len * self.EXPIRE_COEF)
        #        ))
        #    return True
        #return False

    def add_gain(self, gain):
        self.gains.append(gain)
        self.lens.append(len(self)+max(gain,0))

    def add_seed(self, evt, rf, st, en):

        bst, ben = self.blocks[-1]

        self.evts.append(evt)
        self.rsts.append(st)
        self.rens.append(en)

        if (self.rf != rf or max(bst, st) > min(ben, en)):
            self.blocks.append( (st,en) )
            self.add_gain(en-st)

        else:
            l1 = ben-bst
            self.blocks[-1] = (min(st, bst), max(en, ben))
            l2 = self.blocks[-1][1] - self.blocks[-1][0]
            self.add_gain(l2-l1)

        return True

    #TODO add start evt at st-SEED_LEN?
    @property
    def ref_len(self):
        return self.lens[-1]

    @property
    def evt_st(self):
        return self.evts[0]-SEED_LEN

    @property
    def evt_en(self):
        return self.evts[-1]

    def evt_len(self, i=0, j=-1):
        return self.evts[j] - self.evts[i] + 1

    @property
    def evrf_ratio(self, i=0, j=-1):
        return (self.evts[j] - self.evt_st+1) / self.lens[i]

    def __str__(self):
        MAX_REF = 22
        if len(self.rf) < MAX_REF:
            ref = self.rf
        else:
            ref = self.rf[:MAX_REF] + "."

        return "%s:%d-%d (%s)" % (
                ref,
                self.blocks[0][0],
                self.blocks[-1][1],
                "+" if self.fwd else "-")


    #TODO bind confidence to len
    #then make it more complicated
    def __gt__(self, c):
        return c is None or len(self) > len(c)

    def __eq__(self, rhs):
        return self.id == getattr(rhs, "id", None)

    def __len__(self):
        return self.lens[-1]



#!/usr/bin/env python

import sys, os
import numpy as np
import argparse
from uncalled.pafstats import parse_paf
from time import time
from collections import Counter

SAMP_RATE = 4000
CHS = np.arange(512)+1
NCHS = len(CHS)

NTICKS = 32
TICK_MOD = NCHS / NTICKS
PROG = " progress "
SP = '-'*( (NTICKS - len(PROG))//2 - 1)
PROG_HEADER = '|'+SP+PROG+SP+"|\n"

def find_scans(sts,ens,mxs,max_block_gap=1,max_intv_gap=20,min_mux_frac=0.95):

    i = np.argsort(sts)
    sts = sts[i]
    ens = ens[i]
    mxs = mxs[i]

    blocks = list()
    bst = sts[0]
    ben = ens[0]

    for rst,ren in zip(sts[1:], ens[1:]):
        if rst - ben > max_block_gap:
            blocks.append( (bst,ben) )
            bst = rst
            ben = ren
        else:
            ben = max(ren,ben)
            #ben = ren
    blocks.append((bst,ben))

    scan_segs = list()
    scan = list()
    scan_gaps = list()


    prev_en = 0
    for bst,ben in blocks:

        if len(scan) > 0 and bst - scan[-1][1] > max_intv_gap:
            if len(scan) == 4:
                scan_segs.append(scan)
            scan = list()

        mux_counts = Counter(mxs[(sts >= bst) & (sts < ben)])
        mux_counts = [(c,m) for m,c in mux_counts.items()]
        top_count, top_mux = max(mux_counts)

        if top_count / sum((c for c,m in mux_counts)) >= min_mux_frac:

            if top_mux != 4 and len(scan) == 4:
                scan_segs.append(scan)

                scan_gaps.append((gap1,
                                  bst - scan[-1][1]))

                scan = list()

            if len(scan) > 0 and top_mux == len(scan):
                if ben - scan[-1][1] < max_intv_gap:
                    scan[-1] = (scan[-1][0],ben)

                elif top_mux == 1:
                    scan[0] = (bst,ben)
                    gap1 = bst - prev_en

            elif top_mux-1 == len(scan):
                scan.append( (bst,ben) )
                if len(scan) == 1:
                    gap1 = bst - prev_en

            else:
                scan = list()
        else:
            if len(scan) == 4:
                scan_segs.append(scan)
                scan_gaps.append((gap1,
                                  bst - scan[-1][1]))
            scan = list()

        prev_en = ben

    scans = list()

    for segs,gaps in zip(scan_segs, scan_gaps):
        scans.append((segs[0][0]-gaps[0], segs[-1][1]+gaps[1]))

    return scans

class SeqsumProfile:
    PROPS=['chs','sts','lns','mxs',
           'ids','ens','glns','gsts','tms','tds','bps']

    def __init__(self, fname, min_st=0, max_en=np.inf):
        infile = open(fname)
        header = infile.readline().split()
        ch_i = header.index("channel")
        st_i = header.index("start_time")
        ln_i = header.index("duration")
        mx_i = header.index("mux")
        id_i = header.index("read_id")
        tm_i = header.index("template_start")
        td_i = header.index("template_duration")
        bp_i = header.index("sequence_length_template")

        ids=list()
        chs=list()
        sts=list()
        mxs=list() 
        lns=list()
        ens=list()
        tms=list()
        tds=list()
        bps=list()

        SIZE = os.path.getsize(fname)
        MOD = SIZE / NTICKS
        bts = 0
        sys.stderr.write("=")

        for line in infile:
            bts += len(line)
            if bts > MOD:
                bts = 0
                sys.stderr.write("=")
                sys.stderr.flush()

            tabs = line.split()
            st = float(tabs[st_i])
            ln = float(tabs[ln_i])
            en = st+ln
            if st < min_st or en > max_en: continue

            sts.append(st)
            lns.append(ln)
            ens.append(en)
            chs.append(int(tabs[ch_i]))
            mxs.append(int(tabs[mx_i]))
            ids.append(tabs[id_i])
            tms.append(float(tabs[tm_i])-st)
            tds.append(float(tabs[td_i]))
            bps.append(int(tabs[bp_i]))

        sys.stderr.write("\n")
        sys.stderr.write("Procesing run...................\n")

        ids,chs,sts,mxs,lns,ens,tms,tds,bps=map(np.array, (ids,chs,sts,mxs,lns,ens,tms,tds,bps))

        self.ids,self.chs,self.sts,self.mxs,self.lns,self.ens,self.tms,self.tds,self.bps = ids,chs,sts,mxs,lns,ens,tms,tds,bps

        self.sort(np.argsort(sts))

        self.chodr = CHS
        self.chcts = np.array([np.sum(self.chs==ch) 
                               for ch in CHS])

        self.duraiton = np.max(ens)

    def rm_scans(self):
        scans = find_scans(self.sts, self.ens, self.mxs)
        bounds = list()

        sh = 0
        for st,en in (scans):
            m = np.flatnonzero((self.sts+sh >= st) & (self.ens+sh <= en))

            for pr in SeqsumProfile.PROPS:
                a = getattr(self, pr, None)
                if a is not None:
                    setattr(self, pr, np.delete(a,m))

            l = en-st
            bounds.append(st-sh)

            self.sts[self.sts+sh >= st] -= l
            self.ens[self.ens+sh >= st] -= l
            sh += l

        bounds.append(np.max(self.ens))

        self.chcts = np.array([np.sum(self.chs==ch) 
                               for ch in CHS])

        return np.array(bounds)
    
    def compute_eject_delays(self, fname):
        self.dls = np.zeros(len(self.sts))
        self.dls.fill(np.inf)

        idxs = {self.ids[i] : i for i in range(len(self.ids))}
        tlns = self.lns - self.tms

        for p in parse_paf(open(fname)):
            i = idxs.get(p.qr_name, None)
            if i != None and ('ej' in p.tags or 'ub' in p.tags):
                ej = (p.tags['ej'] if 'ej' in p.tags else p.tags['ub'])[0]
                self.dls[i] = max(0, tlns[i] - ((p.qr_len/450.0)+ej))

    def compute_gaps(self):
        self.gsts = np.zeros(len(self.ids))
        self.glns = np.zeros(len(self.ids))

        for ch in CHS:
            cids = self.ids[self.chs==ch]
            csts = self.sts[self.chs==ch]
            cens = self.ens[self.chs==ch]

            gsts = np.insert(cens[:-1], 0, 0)
            glns = csts - gsts

            self.gsts[self.chs==ch] = gsts
            self.glns[self.chs==ch] = glns

    def chsort(self, odr):
        self.chodr = self.chodr[odr]
        self.chcts = self.chcts[odr]


    def sort(self, order=None):
        for pr in SeqsumProfile.PROPS:
            a = getattr(self, pr, None)
            if a is not None:
                setattr(self, pr, a[order])

    def __len__(self):
        return len(self.sts)

def sec_to_samp(sec, coef=1.0):
    return int(np.round(sec*SAMP_RATE*coef))

#def write_itv(out, ch, sc, st_sec, en_sec, time_scale):
#    st_samp = np.round(st_sec*SAMP_RATE*time_scale)
#    en_samp = np.round(en_sec*SAMP_RATE*time_scale)
#    itvs_out.write("%d\t%d\t%d\t%d\n" % (ch,sc,st_samp,en_samp))
#
#def write_gap(out, ch, sc, ln_sec):
#    ln_samp = np.round(ln_sec*SAMP_RATE)
#    out.write("%d\t%d\t%d\n" % (ch, sc, ln_samp))

def load_sim(client, conf):

    t0 = time()

    sys.stderr.write("Loading UNCALLED PAF............\n")
    unc = SeqsumProfile(conf.unc_seqsum)

    unc_scans = unc.rm_scans()
    unc.compute_gaps()

    unc.compute_eject_delays(conf.unc_paf)
    delays = unc.dls[unc.dls != np.inf]
    DELAY = np.median(delays)
    unc.chsort(np.argsort(unc.chcts))

    sys.stderr.write("Generating pattern..............\n")

    ACTIVE_THRESH = np.median(unc.glns)+np.std(unc.glns)

    for ch in CHS:

        if (ch-1) % TICK_MOD == 0:
            sys.stderr.write("=")
            sys.stderr.flush()

        ch_i = unc.chs==ch

        if not np.any(ch_i):
            continue

        gsts = unc.gsts[ch_i]
        glns = unc.glns[ch_i]

        ch_intervals = list()

        #break scan intervals at long gaps
        sc = 0
        itv_st = 0
        for br in np.flatnonzero(glns >= ACTIVE_THRESH):
            act_en = gsts[br] #find first long gap (break)
            #print(ch, itv_st, _en)

            #add all full intervals preceding break
            while unc_scans[sc+1] < act_en:
                itv_en = conf.scan_intv_time

                st_samp = sec_to_samp(itv_st-unc_scans[sc], conf.sim_speed)
                en_samp = sec_to_samp(itv_en, conf.sim_speed)
                client.add_intv(ch,sc,st_samp,en_samp)

                #write_itv(itvs_out, ch, sc, itv_st-unc_scans[sc], itv_en, conf.sim_speed)

                itv_st = unc_scans[sc+1]
                sc += 1

            #add partial intervals before break
            if itv_st != act_en:
                #if np.any((glns < ACTIVE_THRESH) & (gsts > itv_st) & (gsts < act_en)):
                st_samp = sec_to_samp(itv_st-unc_scans[sc], conf.sim_speed)
                en_samp = sec_to_samp(act_en-unc_scans[sc], conf.sim_speed)
                client.add_intv(ch,sc,st_samp,en_samp)

                #write_itv(itvs_out, ch, sc, itv_st-unc_scans[sc], act_en-unc_scans[sc], conf.sim_speed)

            itv_st = act_en + glns[br]
            
            #skip intervals before interval starts
            while unc_scans[sc+1] < itv_st:
                sc += 1

        last = np.max(unc.ens[ch_i]) #time of last read

        #add intervals between last break and final read
        while sc < len(unc_scans)-1 and unc_scans[sc] < last:
            itv_en = min(last - unc_scans[sc], conf.scan_intv_time)

            st_samp = sec_to_samp(itv_st-unc_scans[sc], conf.sim_speed)
            en_samp = sec_to_samp(itv_en, conf.sim_speed)

            client.add_intv(ch,sc,st_samp,en_samp)
            #write_itv(itvs_out,ch,sc,itv_st-unc_scans[sc],itv_en, conf.sim_speed)

            itv_st = unc_scans[sc+1]
            sc += 1

        #write short gaps within each interval
        for sc in range(len(unc_scans)-1):
            sc_i = (gsts > unc_scans[sc]) & ((gsts+glns) <= unc_scans[sc+1])
            for ln in glns[sc_i]:
                if ln < ACTIVE_THRESH and ln > 0:
                    client.add_gap(ch, sc, sec_to_samp(ln))
                    #write_gap(gaps_out, ch, sc, ln)

            for dl in unc.dls[ch_i][sc_i]:
                if dl != np.inf:
                    #write_gap(delays_out, ch, sc, dl)
                    #write_gap(delays_out, ch, sc, DELAY)
                    client.add_delay(ch, sc, sec_to_samp(DELAY))


    sys.stderr.write("\n\n")

    #gaps_out.close()
    #itvs_out.close()
    #delays_out.close()

    sys.stderr.write("Loading control PAF.............\n")
    ctl = SeqsumProfile(conf.ctl_seqsum)
    ctl.rm_scans()
    ctl.chsort(np.argsort(ctl.chcts))

    sys.stderr.write("Ordering reads..................\n")
    #sys.stderr.write(PROG_HEADER)

    #Channels with any reads recieve minimum read count
    min_const = np.zeros(NCHS)
    min_const[unc.chcts > 0] = conf.min_ch_reads

    tgt_total = np.sum(ctl.chcts)

    #Max reads proportional to unc read counts
    max_prpl = tgt_total * unc.chcts / np.sum(unc.chcts)

    #Combine and min const with re-scaled proportional to fit max reads
    remain = max_prpl - min_const
    remain_clp = np.clip(remain, 0, np.inf)
    tgt_counts = min_const + (np.sum(remain) * remain_clp / np.sum(remain_clp))

    #Round and adjust for error
    tgt_counts = np.round(tgt_counts).astype(int)
    dr = -1 if np.sum(tgt_counts) > tgt_total else 1
    i = len(tgt_counts)-1
    while np.sum(tgt_counts) != tgt_total:
        tgt_counts[i] += dr
        i -= 1

    diff = ctl.chcts - tgt_counts

    odr = np.flip(np.argsort(diff),0)
    diff = diff[odr]
    tgt_counts = tgt_counts[odr]
    ctl.chsort(odr)
    unc.chsort(odr)

    sim_reads = [None for c in CHS]
    extra = list()
    e = 0

    for i in range(NCHS):
        if i % TICK_MOD == 0:
            sys.stderr.write("=")
            sys.stderr.flush()
        
        j = ctl.chs==ctl.chodr[i]
        
        ctl_reads = list(zip(ctl.ids[j], ctl.tms[j]))
        new_reads = None

        tgt = tgt_counts[i]

        if diff[i] >= 0:
            new_reads = ctl_reads[:tgt]
            
            if diff[i] > 0:
                extra.append(ctl_reads[tgt:])
        else:
            if e >= len(extra):
                sys.stderr.write("Not enough reads? maybe should haved checked earlier\n")
                sys.exit(1)

            new_reads = ctl_reads
            while len(new_reads) < tgt and e < len(extra):
                needed = tgt-len(new_reads)
                if len(extra[e]) > needed:
                    new_reads += extra[e][:needed]
                    extra[e] = extra[e][needed:]
                else:
                    new_reads += extra[e]
                    e += 1

            if len(new_reads) < tgt:
                sys.stderr.write("Not enough reads? again? not sure I should be here\n")
                sys.exit(1)

        sim_reads[unc.chodr[i]-1] = new_reads

    sys.stderr.write("\n")

    for ch in CHS:
        for rd,tm in sim_reads[ch-1]:
            client.add_read(ch, rd, sec_to_samp(tm))
            #reads_out.write("%3d %s %d\n" % (ch, rd, np.round(tm*SAMP_RATE)))
            #reads_out.write("%3d %s %d\n" % (ch, rd, 0))
    #reads_out.close()



import traceback
from threading import Event, Thread
import sys
import time
import numpy as np
import logging

try:
    import read_until
    ru_loaded = True

except ImportError:
    ru_loaded = False

if ru_loaded:

    MK_ERROR_STATUS = 0
    MK_READY = 1
    MK_STARTING = 2
    MK_PROCESSING = 3
    MK_FINISHING = 4

    RN_STARTING = 0
    RN_RUNNING = 1
    RN_FINISHING = 2
    RN_COMPLETED = 3

    class Client(read_until.ReadUntilClient):
        def __init__(self, 
                     mk_host=8000, 
                     mk_port="127.0.0.1", 
                     chunk_size=1.0, 
                     scan_thresh=0.99,
                     num_channels=512):

            logging.basicConfig(
                format='[%(asctime)s - %(name)s] %(message)s',
                datefmt='%H:%M:%S', level=logging.INFO)

            self.chmon_running = Event()
            self.chmon_thread = None

            read_until.ReadUntilClient.__init__(
                    self,
                    mk_host=mk_host,  
                    mk_port=mk_port,  
                    one_chunk=False,    
                    filter_strands=True,
                    prefilter_classes={'strand', 'strand1', 'strand2'}
            )


            self.in_scan = True 
            self.scan_thresh = scan_thresh
            self.chunk_size = chunk_size
            self.num_channels = num_channels

            self.ch_mux = np.zeros(num_channels, dtype=int)
            self.mux_counts = np.zeros(5, dtype=float)
            self.mux_counts[0] = len(self.ch_mux)

            self.logger.setLevel(logging.WARNING)
            self.unc_log = logging.getLogger('UNCALLED')

            #self.device = minknow.Device(self.connection)


        def run(self, steady_wait=10, scan_wait=5, refresh=0.5, **kwargs):
            self.anl_client = self.connection.analysis_configuration

            if not self._wait_for_start(steady_wait, refresh):
                return False

            self._start_chmon()

            read_until.ReadUntilClient.run(self, last_channel=self.num_channels, *kwargs)

            self.start_time = time.time()
            
            return True

        def reset(self):
            read_until.ReadUntilClient.reset(self)

            if self.chmon_thread is not None:
                self.chmon_running.clear()
                self.chmon_thread.join()
                if self.chmon_thread.is_alive():
                    self.unc_log.error("Channel monitor didn't exit right")
            self.chmon_thread = None

        def get_runtime(self):
            return time.time() - self.start_time

        def should_eject(self):
            return not self.in_scan
        
        def log(self, msg):
            self.unc_log.info(msg)

        def _start_chmon(self):
            self.start_time = time.time()
            self.last_scan = time.time()
            self.in_scan = True

            self.chmon_running.set()
            self.chmon_thread = Thread(
                target=self._run_chmon,
                name="channel_monitor"
            )
            self.chmon_thread.start()

        def _run_chmon(self, **kwargs):
            channels = self.connection.data.get_channel_states(
                first_channel = 1,
                last_channel = self.num_channels,
                use_channel_states_ids = False
            )

            try:
                self._update_muxs(channels)
            except Exception as e:
                self.unc_log.error(traceback.format_exc())

            # signal to the server that we are done with the stream.
            channels.cancel()

            self.log("Finished")

        def _scan_update(self):
            #ps = self.in_scan

            m = np.argmax(self.mux_counts)
            self.in_scan = (m != 0 and 
                            (self.mux_counts[m] / (len(self.ch_mux)-self.mux_counts[0])
                                > self.scan_thresh))

            if self.in_scan:
                self.last_scan = time.time()

            #if self.in_scan != ps:
            #    sys.stderr.write("%.2f in_scan changed to %d\n" % 
            #                     (self.get_runtime(), self.in_scan))

        def _update_muxs(self, channels):
            for channels in channels:
                if not self.chmon_running.is_set():
                    break

                for ch in channels.channel_states:
                    nmx = int(ch.config.well)
                    omx = self.ch_mux[ch.channel-1]
                    if omx != nmx:
                        self.mux_counts[omx] -= 1
                        self.mux_counts[nmx] += 1
                        self.ch_mux[ch.channel-1] = nmx

                self._scan_update()

                if self._get_minknow_status() != MK_PROCESSING:
                    self.chmon_running.clear()
                    self.running.clear()
                    break



        def _get_minknow_status(self):
            return self.connection.acquisition.current_status().status

        def _get_run_state(self):
            return self.connection.acquisition.get_acquisition_info().state


        def _wait_for_start(self, steady=10, refresh=0.01):
            if self._get_minknow_status() == MK_PROCESSING:
                self.log("Run already in progress")
                if self._update_chunk_len(False):
                    self.unc_log.error("ERROR: cannot set chunk size when run is in progress. Please stop the sequencing run and start UNCALLED before re-starting if you want to change the chunks size.")
                    return False
                return True

            processing = False
            proc_time = None
            
            self.log("Waiting for run to start")

            while True:
                status = self._get_minknow_status()
                state = self._get_run_state()

                if status == MK_STARTING or state == RN_STARTING:
                    self._update_chunk_len(True)

                if status == MK_PROCESSING:
                    if proc_time == None:
                        proc_time = time.time()

                        if not processing:
                            self.log("Waiting for steady state")
                            processing = True

                    elif time.time() - proc_time >= steady:
                        return True

                elif proc_time != None:
                    proc_time = None

                time.sleep(refresh)

            return False

        def _update_chunk_len(self, change=True):
            try:
                anl_config = self.anl_client.get_analysis_configuration()
            except:
                self.unc_log.warning("Warning: failed to check chunk size. If chunk size is set to 1 second this is fine, otherwise stop the run, use a 1 second chunk size, and please report this error.")
                return False

                if self.chunk_size != anl_config.read_detection.break_reads_after_seconds.value:
                    if change:
                        #Credit to Matt Loose for directing me to the parameter to change
                        anl_config.read_detection.break_reads_after_seconds.value=self.chunk_size
                        self.anl_client.set_analysis_configuration(anl_config)
                        self.log("Setting chunk size to %.2f" % (self.chunk_size))
                    return True
            return False





#!/usr/bin/env python

import sys, os
import numpy as np
import matplotlib.pyplot as plt
import argparse

def mask_seq(seq, kmer):
    ranges = list()
    
    k = len(kmer)
    n = 0

    useq = seq.upper()

    i = useq.find(args.kmer)
    while i >= 0:
        j = i + k
        if len(ranges) == 0 or i > ranges[-1][1]:
            ranges.append( (i,j) )
        else:
            #sys.stderr.write(str(ranges[-1]) + " ")
            ranges[-1] = (ranges[-1][0],j)
            #sys.stderr.write(str(ranges[-1]) + "\n")

        n += 1
        i = useq.find(args.kmer, i+1)

    masked = list()
    i = 0
    for j,k in ranges:
        masked.append(seq[i:j] + "N"*(k-j))
        i = k

    if len(ranges) > 0:
        masked.append(seq[ranges[-1][1]:])
        return "".join(masked), n


    return seq, 0

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("infile", type=str, default=None)
    parser.add_argument("-k", "--kmer", required=False, type=str, default=None)
    parser.add_argument("-o", "--outfile", required=False, type=str, default=None)
    args = parser.parse_args(sys.argv[1:])

    run = 0
    prev = None

    seq = list()

    masked = 0
    headers = list()
    masked_seqs = list()

    k=len(args.kmer)

    Ns = "N"*k


    for line in open(args.infile):
        if line[0] == ">":
            if len(seq) > 0:
                m,n = mask_seq("".join(seq), args.kmer)
                masked_seqs.append(m)
                masked += n
                seq = list()
            headers.append(line.strip())
        else:
            seq.append(line.strip())

    m,n = mask_seq("".join(seq), args.kmer)
    masked_seqs.append(m)
    masked += n

    sys.stderr.write("masked %d occurences of %s\n" % (masked, args.kmer))

    for h,s in zip(headers, masked_seqs):
        print(h)
        print(s)



#!/usr/bin/env python

# MIT License
#
# Copyright (c) 2018 Sam Kovaka <skovaka@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import division
import sys                         
import os
import numpy as np
import argparse
#from uncalled import mapping, params
import uncalled as unc
from bisect import bisect_left, bisect_right

UNCL_SUFF = ".uncl"
AMB_SUFF = ".amb"
ANN_SUFF = ".ann"
BWT_SUFF = ".bwt"
PAC_SUFF = ".pac" 
SA_SUFF = ".sa"
BWA_SUFFS = [AMB_SUFF, ANN_SUFF, BWT_SUFF, PAC_SUFF, SA_SUFF]

ROOT_DIR = os.path.dirname(os.path.realpath(__file__))

MODEL_FNAME = os.path.join(ROOT_DIR, "conf/r94_5mers.txt")
CONF_DEFAULTS = os.path.join(ROOT_DIR, "conf/defaults.toml")

def power_fn(xmax, ymin, ymax, exp, N=100):
    dt = 1.0/N
    t = np.arange(0, 1+dt, dt)

    return t*xmax, (t**exp) * (ymax-ymin) + ymin

class IndexParameterizer:
    MODEL_THRESHS_FNAME = os.path.join(ROOT_DIR, "conf/r94_5mers_threshs.txt")

    def __init__(self, args):
        self.out_fname = args.bwa_prefix + UNCL_SUFF

        self.pck1 = args.matchpr1
        self.pck2 = args.matchpr2

        self.calc_map_stats(args)
        self.get_model_threshs()

        self.functions = dict()

    def calc_map_stats(self, args):

        ann_in = open(args.bwa_prefix + ANN_SUFF)
        header = ann_in.readline()
        ref_len = int(header.split()[0])
        ann_in.close()

        approx_samps = ref_len / args.max_sample_dist
        if approx_samps < args.min_samples:
            sample_dist = int(np.ceil(ref_len/args.min_samples))
        elif approx_samps > args.max_samples:
            sample_dist = int(np.floor(ref_len/args.max_samples))
        else:
            sample_dist = args.max_sample_dist

        fmlens = unc.self_align(args.bwa_prefix, sample_dist)
        path_kfmlens = [p[args.kmer_len-1:] if len(p) >= args.kmer_len else [1] for p in fmlens]

        max_pathlen = 0
        all_pathlens = [len(p) for p in path_kfmlens if len(p) <= args.max_replen]
        gt1_counts = np.zeros(max(all_pathlens))
        for l in all_pathlens:
            for i in range(l):
                gt1_counts[i] += 1

        max_pathlen = np.flatnonzero(gt1_counts / len(all_pathlens) <= args.pathlen_percentile)[0]
        max_fmexp = int(np.log2(max([p[0] for p in path_kfmlens])))+1
        fm_path_mat = np.zeros((max_fmexp, max_pathlen))

        for p in path_kfmlens:
            for i in range(min(max_pathlen, len(p))):
                fm_path_mat[int(np.log2(p[i])), i] += 1
            for i in range(len(p), max_pathlen):
                fm_path_mat[0, i] += 1

        mean_fm_locs = list()
        for f in range(max_fmexp):
            loc_weights = fm_path_mat[f] / np.sum(fm_path_mat[f])
            mean_fm_locs.append(np.sum([i*loc_weights[i] for i in range(max_pathlen)]))
        self.fm_locs = np.array(mean_fm_locs)

        mean_loc_fms = list()
        for p in range(max_pathlen):
            fm_weights = fm_path_mat[:,p] / np.sum(fm_path_mat[:,p])
            mean_loc_fms.append(np.sum([i*fm_weights[i] for i in range(max_fmexp)]))
        self.loc_fms = np.array(mean_loc_fms)

        self.speed_denom = np.sum(self.loc_fms)

        self.conf_locs = np.arange(np.round(self.fm_locs[0]))
        self.all_locs = np.arange(max_pathlen)

    def get_model_threshs(self, fname=MODEL_THRESHS_FNAME):
        prob_thresh_in = open(fname)
        threshs = list()
        freqs = list()
        counts = list()
        for line in prob_thresh_in:
            thresh, freq, count = line.split()
            threshs.append(float(thresh))
            freqs.append(float(freq))
            counts.append(float(count))

        self.model_ekms = np.flip(np.array(threshs),0)
        self.model_pcks = np.flip(np.array(freqs),0)
        self.model_counts = np.flip(np.array(counts),0)

    def get_fn_speed(self, fn_locs, fn_pcks):
        pcks = np.interp(self.all_locs, fn_locs, fn_pcks)
        counts = np.interp(pcks, self.model_pcks, self.model_counts)
        speed = np.dot(counts, self.loc_fms) / (self.speed_denom)
        return speed

    def get_fn_prob(self, fn_locs, fn_pcks):
        return np.prod(np.interp(self.conf_locs, fn_locs, fn_pcks))

    def add_preset(self, name, tgt_prob=None, tgt_speed=None, exp_st=2, init_fac=2, eps=0.00001):

        exp = exp_st
        exp_min, exp_max = (None, None)

        pdelta = None

        pck1 = self.pck1
        pck2 = self.pck2

        sys.stderr.write("Computing %s parameters\n" % name)

        while True:
            fn_locs,fn_pcks = power_fn(self.fm_locs[0], pck1, pck2, exp)

            if tgt_prob is not None:
                delta = self.get_fn_prob(fn_locs, fn_pcks) - tgt_prob
            elif tgt_speed is not None:
                delta = self.get_fn_speed(fn_locs, fn_pcks) - tgt_speed

            if abs(delta) <= eps:
                break
            
            if delta == pdelta:
                #This works well for small references
                #TODO: check for larger references
                sys.stderr.write("Maxed out %s parameters\n" % name)
                break
            pdelta = delta

            if delta < 0:
                exp_max = exp
            else:
                exp_min = exp
            
            pexp = exp

            if exp_max == None:
                exp *= init_fac
            elif exp_min == None:
                exp /= init_fac
            else:
                exp = exp_min + ((exp_max - exp_min) / 2.0)

            #for floating point rounding errors
            if exp == pexp:
                break

        fm_pcks = np.interp(self.fm_locs, fn_locs, fn_pcks)
        fm_ekms = np.interp(fm_pcks, self.model_pcks, self.model_ekms)
        prob = self.get_fn_prob(fn_locs, fn_pcks)
        speed = self.get_fn_speed(fn_locs, fn_pcks)

        #while len(fm_ekms) > 2 and fm_ekms[-1] == fm_ekms[-2]:
        #    fm_ekms = fm_ekms[:-1]

        sys.stderr.write("Writing %s parameters\n" % name)
        self.functions[name] = (fm_ekms, prob, speed)

    def write(self):
        params_out = open(self.out_fname, "w")
        
        for name, fn in self.functions.items():
            ekms, prob, speed = fn
            params_out.write("%s\t%s\t%.5f\t%.3f\n" % (name, ",".join(map(str,ekms)), prob, speed))

        params_out.close()




from setuptools import setup, find_packages, Extension
from setuptools.command.build_ext import build_ext
import os
import subprocess
import sys

__version__ = "2.2"

ROOT_DIR = os.getcwd()

SUBMOD_DIR = os.path.join(ROOT_DIR, "submods")
SUBMODS = [
    "bwa", 
    "fast5", 
    "hdf5", 
    "pdqsort", 
    "toml11"
]

class get_pybind_include(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked. """

    def __str__(self):
        import pybind11
        return pybind11.get_include()

class pre_build(build_ext):
    def run(self):

        submods_loaded = True
        for submod in SUBMODS:
            if not os.path.exists(os.path.join(SUBMOD_DIR, submod)):
                submods_loaded = False
                break

        if not submods_loaded:
            sys.stderr.write("Downloading submodules\n")
            subprocess.check_call([
                "git", "submodule", "update", "--init"
            ])
        else:
            sys.stderr.write("All submodules present\n")

        if os.path.exists("./submods/bwa/libbwa.a"):
            sys.stderr.write("Found libbwa.a\n")
        else:
            sys.stderr.write("building libbwa\n")

            subprocess.check_call([
                "make", 
                 "-C", "./submods/bwa", 
                 "-f", "../../src/Makefile_bwa"
            ])

        if os.path.exists("./submods/hdf5/lib/libhdf5.a"):
            sys.stderr.write("Found libhdf5.a\n")
        else:

            hdf5_dir = os.path.join(ROOT_DIR, "submods/hdf5")

            os.chdir(hdf5_dir)

            subprocess.check_call([
                "./configure", 
                    "--enable-threadsafe", 
                    "--disable-hl",
                    "--prefix", hdf5_dir,
                    "--enable-shared=no",
                    "--with-pic=yes"
            ])

            subprocess.check_call(["make"])
            subprocess.check_call(["make", "install"])

            os.chdir(ROOT_DIR)

        build_ext.run(self)

uncalled = Extension(
    "_uncalled",

    sources = [
       "src/event_profiler.cpp", 
       "src/pybinder.cpp",
       "src/client_sim.cpp",
       "src/fast5_reader.cpp",
       "src/mapper.cpp",
       "src/self_align_ref.cpp",
       "src/map_pool.cpp",
       "src/event_detector.cpp", 
       "src/read_buffer.cpp",
       "src/chunk.cpp",
       "src/realtime_pool.cpp",
       "src/seed_tracker.cpp", 
       "src/normalizer.cpp", 
       "src/range.cpp"
    ],

    include_dirs = [
        "./submods", #TODO: consistent incl paths?
        "./submods/hdf5/include", 
        "./submods/fast5/include",
        "./submods/pdqsort",
        "./submods/toml11",
        get_pybind_include()
    ],

    library_dirs = [
        "./submods/bwa", 
        "./submods/hdf5/lib"
    ],

    libraries = ["bwa", "hdf5", "z", "dl", "m"],

    extra_compile_args = ["-std=c++11", "-O3"],

    define_macros = [("PYBIND", None)]
)

setup(
    name = "uncalled",
    version = __version__,
    description = "Rapidly maps raw nanopore signal to DNA references",
    author = "Sam Kovaka",
    author_email = "skovaka@gmail.com",
    url = "https://github.com/skovaka/UNCALLED",

    classifiers=[
      "Programming Language :: Python :: 3"
    ],

    python_requires=">=3.6",

    setup_requires=[
        'pybind11>=2.5.0', 
        'read-until==3.0.0'
    ],

    packages=find_packages(),
    include_package_data=True,
    scripts = ['scripts/uncalled'],
    ext_modules = [uncalled],
    cmdclass={'build_ext': pre_build},
)



#!/usr/bin/env python

import sys
import numpy as np
import re
import argparse

class PafEntry:
    def __init__(self, line, tags=None):
        fromstr = type(line) != list
        if fromstr:
            tabs = line.split()
        else:
            tabs = line

        self.qr_name = tabs[0]
        self.qr_len = int(tabs[1])
        self.is_mapped = tabs[4] != ("*" if fromstr else None)

        if self.is_mapped:
            self.qr_st = int(tabs[2])
            self.qr_en = int(tabs[3])
            self.is_fwd = tabs[4] == ('+' if fromstr else True)
            self.rf_name = tabs[5]
            self.rf_len = int(tabs[6])
            self.rf_st = int(tabs[7])
            self.rf_en = int(tabs[8])
            self.match_num = int(tabs[9])
            self.aln_len = int(tabs[10])
            self.qual = int(tabs[11])
        else:
            self.qr_st = 1
            self.qr_en = self.qr_len
            self.is_fwd=self.rf_name=self.rf_len=self.rf_st=self.rf_en=self.match_num=self.aln_len=self.qual=None

        self.tags = dict() if tags==None else tags 
        for k,t,v in (s.split(":") for s in tabs[12:]):
            if t == 'f':
                v = float(v)
            elif t == 'i':
                v = int(v)
            elif t not in ['A', 'Z','B', 'H']:
                sys.stderr.write("Error: invalid tag type \"%s\"\n" % t)
                sys.exit(1)
            self.tags[k] = (v,t)

    def rev(self):
        return PafEntry( [self.rf_name, self.rf_len, self.rf_st, self.rf_en, self.is_fwd, 
                          self.qr_name, self.qr_len, self.qr_st, self.qr_en, self.match_num,
                          self.aln_len, self.qual], self.tags )

    def get_tag(self, k):
        return self.tags.get(k, (None,None))[0]
    
    def set_tag(self, k, v, t=None):
        if t == None:
            if isinstance(v, int):
                t = 'i'
            elif isinstance(v, float):
                t = 'f'
            else:
                t = 'Z'
        self.tags[k] = (v,t)

    def qry_loc(self):
        return (self.qr_name, self.qr_st, self.qr_en)

    def ref_loc(self):
        return (self.rf_name, self.rf_st, self.rf_en)

    def ext_ref(self, ext=1.0):
        st_shift = int(self.qr_st*ext)
        en_shift = int((self.qr_len - self.qr_en)*ext)

        if self.is_fwd:
            return (max(1, self.rf_st - st_shift),
                    min(self.rf_len, self.rf_en + en_shift))
        else:
            return (max(1, self.rf_st - en_shift),
                    min(self.rf_len, self.rf_en + st_shift))
    
    def overlaps(self, paf2, ext=0.0):
        st1, en1 = self.ext_ref(ext)
        st2, en2 = paf2.ext_ref(ext)
        return (self.is_mapped and paf2.is_mapped and self.rf_name.startswith(paf2.rf_name) and 
                max(st1, st2) <= min(en1, en2))

    def contains(self, paf2):
        return (self.is_mapped and paf2.is_mapped and self.rf_name.startswith(paf2.rf_name) and 
                self.rf_st <= paf2.rf_st and self.rf_en >= paf2.rf_en)

    def __lt__(self, paf2):
        return self.qr_name < self.qr_name

    def __str__(self):
        tagstr = "\t".join( (":".join([k,v[1],str(v[0])]) for k,v in self.tags.items()))
        if self.is_mapped:
            s = "%s\t%d\t%d\t%d\t%s\t%s\t%d\t%d\t%d\t%d\t%d\t%d\t%s" % (
                 self.qr_name, self.qr_len, self.qr_st, self.qr_en, 
                 '+' if self.is_fwd else '-', self.rf_name, 
                 self.rf_len, self.rf_st, self.rf_en, 
                 self.match_num, self.aln_len, self.qual, tagstr)
        else:
            s = "\t".join((self.qr_name,str(self.qr_len)) + ("*",)*10 + (tagstr,))

        return s


def parse_paf(infile, max_load=None):
    if isinstance(infile, str):
        infile = open(infile)
    c = 0
    for l in infile:
        if l[0] == "#": continue
        if max_load != None and c >= max_load: break
        yield PafEntry(l)
        c += 1

def paf_ref_compare(qry, ref, ret_qry=True, check_locs=True, ext=1.5):
    if type(ref) == dict:
        ref_locs = ref
    else:
        ref_locs = dict()
        for r in ref:
            l = ref_locs.get(r.qr_name, None)
            if l == None:
                ref_locs[r.qr_name] = [r]
            else:
                l.append(r)

    tp = list()
    tn = list()
    fp = list()
    fn = list()
    fp_unmap = list()

    for q in qry:
        rs = ref_locs.get(q.qr_name, [None])
        if q.is_mapped:
            if rs == [None] or not rs[0].is_mapped:
                fp_unmap.append(q if ret_qry else rs[0])
                continue
            match = False
            for r in rs:
                if ((check_locs and q.overlaps(r, ext)) or 
                    (not check_locs and q.rf_name == r.rf_name)):
                        match = True
                        tp.append(q if ret_qry else r)
                        break
            if not match:
                fp.append(q if ret_qry else r)
        else:
            if rs == [None] or not rs[0].is_mapped:
                tn.append(q if ret_qry else rs[0])
            else:
                fn.append(q if ret_qry else rs[0])

    return tp, tn, fp, fn, fp_unmap

def add_opts(parser):
    parser.add_argument("infile", type=str, help="PAF file output by UNCALLED")
    parser.add_argument("-n", "--max-reads", required=False, type=int, default=None, help="Will only look at first n reads if specified")
    parser.add_argument("-r", "--ref-paf", required=False, type=str, default=None, help="Reference PAF file. Will output percent true/false positives/negatives with respect to reference. Reads not mapped in reference PAF will be classified as NA.")
    parser.add_argument("-a", "--annotate", action='store_true', help="Should be used with --ref-paf. Will output an annotated version of the input with T/P F/P specified in an 'rf' tag")

def run(args):
    locs = [p for p in parse_paf(args.infile, args.max_reads)]

    num_mapped = sum([p.is_mapped for p in locs])

    statsout = sys.stderr if args.annotate else sys.stdout

    statsout.write("Summary: %d reads, %d mapped (%.2f%%)\n\n" % (len(locs), num_mapped, 100*num_mapped/len(locs)))

    if args.ref_paf != None:
        statsout.write("Comparing to reference PAF\n")
        tp, tn, fp, fn, fp_unmap = paf_ref_compare(locs, parse_paf(args.ref_paf))
        ntp,ntn,nfp,nfn,nfp_unmap = map(len, [tp, tn, fp, fn, fp_unmap])
        n = len(locs)

        statsout.write("     P     N\n")
        statsout.write("T %6.2f %5.2f\n" % (100*ntp/n, 100*ntn/n))
        statsout.write("F %6.2f %5.2f\n" % (100*(nfp)/n, 100*nfn/n))
        statsout.write("NA: %.2f\n\n" % (100*nfp_unmap/n))

        if args.annotate:
            group_labels = [(tp, "tp"), 
                            (tn, "tn"), 
                            (fp, "fp"),
                            (fn, "fn"),
                            (fp_unmap, "na")]

            for grp,lab in group_labels:
                for p in grp:
                    p.set_tag("rf", lab, "Z")
                    sys.stdout.write("%s\n" % p)

    if locs[0].get_tag('mt') != None:
        map_ms = np.array([p.get_tag('mt') for p in locs if p.is_mapped])
        map_bp = np.array([p.qr_en for p in locs if p.is_mapped])
        map_bpps = 1000*map_bp/map_ms

        statsout.write("Speed            Mean    Median\n")
        statsout.write("BP per sec: %9.2f %9.2f\n" % (np.mean(map_bpps), np.median(map_bpps)))
        statsout.write("BP mapped:  %9.2f %9.2f\n" % (np.mean(map_bp),   np.median(map_bp)))
        statsout.write("MS to map:  %9.2f %9.2f\n" % (np.mean(map_ms),   np.median(map_ms)))
